
// Blockbuster
// Avisynth filter for eliminating DCT blocks
//
// By Ross Thomas <ross@grinfinity.com>
//
// There is no copyright on this code, and there are no conditions
// on its distribution or use. Do with it what you will.

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <assert.h>

#ifdef AVISYNTH_25
#include "avisynth25.h"
#else
#include "avisynth.h"
#endif

#include "ShowProcessor.h"
#include "NoiseProcessor.h"
#include "DitherProcessor.h"
#include "SharpenProcessor.h"
#include "BlurProcessor.h"

class Blockbuster : public GenericVideoFilter
{
public:
	Blockbuster(PClip clip, const char* method, int block_size,
		int detail_min, int detail_max, int luma_offset,
		int luma_threshold, double mean, double variance,
		int cache, int seed, int strength, IScriptEnvironment* env);
	virtual ~Blockbuster();

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);

private:
	Processor* processor;
};

Blockbuster::Blockbuster(PClip clip, const char* method, int block_size,
						 int detail_min, int detail_max,
						 int luma_offset, int luma_threshold,
						 double mean, double variance, int cache, int seed,
						 int strength, IScriptEnvironment* env)
	: GenericVideoFilter(clip)
{
#ifdef AVISYNTH_25
	if(!vi.IsYV12() && !vi.IsYUY2())
		env->ThrowError("Blockbuster: Clip must be in YUY2 or YV12 format");
#else
	if(!vi.IsYUY2())
		env->ThrowError("Blockbuster: Clip must be in YUY2 format");
#endif

	if(!stricmp(method, "show"))
	{
		processor = new ShowProcessor(block_size, detail_min,
			detail_max, luma_offset, luma_threshold, env);
	} else if(!stricmp(method, "noise"))
	{
		processor = new NoiseProcessor(block_size, detail_min,
			detail_max, luma_offset, luma_threshold, mean, variance,
			cache, seed, env);
	} else if(!stricmp(method, "dither"))
	{
		processor = new DitherProcessor(block_size, detail_min,
			detail_max, luma_offset, luma_threshold, vi.width,
			vi.height, mean, variance, seed, env);
	} else if(!stricmp(method, "sharpen"))
	{
		processor = new SharpenProcessor(block_size, detail_min,
			detail_max, luma_offset, luma_threshold, strength, env);
	} else if(!stricmp(method, "blur"))
	{
		processor = new BlurProcessor(block_size, detail_min,
			detail_max, luma_offset, luma_threshold, strength, env);
	} else
	{
		env->ThrowError("Blockbuster: Method must be \"show\", \"noise\", "
			"\"dither\", \"sharpen\" or \"blur\"");
	}
}

Blockbuster::~Blockbuster()
{
	delete processor;
}

PVideoFrame Blockbuster::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame frame = child->GetFrame(n, env);
	env->MakeWritable(&frame);

#ifdef AVISYNTH_25
	if(vi.IsYV12())
	{
		processor->DoFrame_YV12(frame->GetWritePtr(PLANAR_Y),
			frame->GetPitch(PLANAR_Y), frame->GetRowSize(PLANAR_Y),
			frame->GetHeight(PLANAR_Y), env);
	} else
	{
		processor->DoFrame_YUY2(frame->GetWritePtr(),
			frame->GetPitch(), frame->GetRowSize(),
			frame->GetHeight(), env);
	}
#else
	processor->DoFrame_YUY2(frame->GetWritePtr(), frame->GetPitch(),
		frame->GetRowSize(), frame->GetHeight(), env);
#endif

	return frame;
}

AVSValue __cdecl Create_Blockbuster(AVSValue args, void* /* user_data */,
									IScriptEnvironment* env )
{
	enum ARGS {
		CLIP,
		METHOD, BLOCK_SIZE, DETAIL_MIN, DETAIL_MAX, 
			LUMA_OFFSET, LUMA_THRESHOLD,					// Common
		MEAN, VARIANCE, CACHE, SEED,						// Noise and dither
		STRENGTH											// Sharpen and blur
	};

	// Check correct parameters are used with each method

	const char* method = args[METHOD].AsString("show");

	if(stricmp(method, "show")) // Show mode just passes through
	{
		if(((!stricmp(method, "noise") || !stricmp(method, "dither")) &&
			args[STRENGTH].Defined()) ||
		   ((!stricmp(method, "sharpen") || !stricmp(method, "blur")) &&
		     (args[MEAN].Defined() || args[VARIANCE].Defined() ||
			  args[CACHE].Defined() || args[SEED].Defined())))
		{
			env->ThrowError("Blockbuster: Invalid parameter for method");
		}
	}

	return new Blockbuster(
		args[CLIP].AsClip(),
		method,
		args[BLOCK_SIZE].AsInt(8),
		args[DETAIL_MIN].AsInt(1),
		args[DETAIL_MAX].AsInt(10),
		args[LUMA_OFFSET].AsInt(0),
		args[LUMA_THRESHOLD].AsInt(25),
		args[MEAN].AsFloat(0),
		args[VARIANCE].AsFloat(1),
		args[CACHE].AsInt(256) * 1024,
		args[SEED].AsInt(0),
		args[STRENGTH].AsInt(25),
		env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit(IScriptEnvironment* env)
{
	env->AddFunction("Blockbuster",
		"c"
		"[method]s[block_size]i[detail_min]i[detail_max]i"
			"[luma_offset]i[luma_threshold]i"
		"[mean]f[variance]f[cache]i[seed]i"
		"[strength]i",
		Create_Blockbuster, 0);
	return "Blockbuster: Eliminate DCT blocks";
}
