
// Blockbuster
// Avisynth filter for eliminating DCT blocks
//
// By Ross Thomas <ross@grinfinity.com>
//
// There is no copyright on this code, and there are no conditions
// on its distribution or use. Do with it what you will.

// This method applies a simple blur matrix:
//
// 1 1 1
// 1 1 1
// 1 1 1 
//
// to the entire block, except for the edges which use modified
// matrices. The blurred result is then weighted with the orignal
// based on the strength setting provided.

#include "Processor.h"

class BlurProcessor : public Processor
{
public:
	BlurProcessor(int block_size, int detail_min,
		int detail_max, int luma_offset, int luma_threshold,
		int strength, IScriptEnvironment* env);
	virtual ~BlurProcessor();

protected:
	virtual void Reset() {}

	virtual void ProcessBlock_YUY2(unsigned char* p, int pitch, int row_size,
		int height, int block_x, int block_y, IScriptEnvironment* env);

#ifdef AVISYNTH_25
	virtual void ProcessBlock_YV12(unsigned char* p, int pitch, int row_size,
		int height, int block_x, int block_y, IScriptEnvironment* env);
#endif

private:
	unsigned char blur_lookup[256], orig_lookup[256];
	short* dest;
};

BlurProcessor::BlurProcessor(int block_size, int detail_min,
							 int detail_max,
							 int luma_offset, int luma_threshold,
							 int strength, IScriptEnvironment* env)
	: Processor(block_size, detail_min, detail_max,
				luma_offset, luma_threshold, env)
{
	if(strength < 1 || strength > 100)
		env->ThrowError("Blockbuster: Strength must be >= 1 and <= 100");

	// Generate strength lookup tables

	float blur_adj = strength / 100.0f, orig_adj = (100 - strength) / 100.0f;
	for(int i = 0; i < 256; ++i)
	{
		blur_lookup[i] = unsigned char(i * blur_adj + 0.5f);
		orig_lookup[i] = unsigned char(i * orig_adj + 0.5f);
	}

	// Temp buffer

	dest = new short[block_size * block_size];
}

BlurProcessor::~BlurProcessor()
{
	delete [] dest;
}

void BlurProcessor::ProcessBlock_YUY2(unsigned char* p, int pitch, int row_size,
									  int height, int /* block_x */, int /* block_y */,
									  IScriptEnvironment* /* env */)
{
	unsigned char* pp = p;
	short* dp = dest;

	const int w = row_size >> 1;

	const int h = height - 2, rs = w - 2, diff = pitch - row_size;

	// Apply [ 0 0 0, 0 1 1, 0 1 1 ] to top-left corner

	*dp++ = (p[0] + p[2] + p[pitch] + p[2 + pitch]) >> 2;
	p += 2;

	// Apply [ 0 0 0, 1 1 1, 1 1 1 ] to top row

	for(int x = 0; x < rs; ++x)
	{
		*dp++ = (p[-2] + p[0] + p[2] + p[pitch - 2] + p[pitch] + p[pitch + 2]) / 6;
		p += 2;
	}

	// Apply [ 0 0 0, 1 1 0, 1 1 0 ] to top-right corner

	*dp++ = (p[-2] + p[0] + p[pitch - 2] + p[pitch]) >> 2;
	p += 2;

	// Skip "leftover"

	p += diff;

	// y=1 to y=height-1

	for(int y = 0; y < h; ++y)
	{
		// Apply [ 0 1 1, 0 1 1, 0 1 1 ] to left edge

		*dp++ = (p[-pitch] + p[2 - pitch] + p[0] + p[2] + p[pitch] + p[2 + pitch]) / 6;
		p += 2;

		// Apply [ 1 1 1, 1 1 1, 1 1 1 ]

		for(int x = 0; x < rs; ++x)
		{
			*dp++ = (p[-2 - pitch] + p[-pitch] + p[2 - pitch] + p[-2] + p[0] +
				p[2] + p[pitch - 2] + p[pitch] + p[pitch + 2]) / 9;
			p += 2;
		}

		// Apply [ 1 1 0, 1 1 0, 1 1 0 ] to right edge

		*dp++ = (p[-2 - pitch] + p[-pitch] + p[-2] + p[0] +
			p[pitch - 2] + p[pitch]) / 6;
		p += 2;

		// Skip "leftover"

		p += diff;
	}

	// Apply [ 0 1 1, 0 1 1, 0 0 0 ] to bottom-left corner

	*dp++ = (p[-pitch] + p[2 - pitch] + p[0] + p[2]) >> 2;
	p += 2;

	// Apply [ 1 1 1, 1 1 1, 0 0 0 ] to bottom row

	for(int x = 0; x < rs; ++x)
	{
		*dp++ = (p[-2 - pitch] + p[-pitch] + p[2 - pitch] +
			p[-2] + p[0] + p[2]) / 6;
		p += 2;
	}

	// Apply [ 1 1 0, 1 1 0, 0 0 0 ] to bottom-right corner

	dp[0] = (p[-2 - pitch] + p[-pitch] + p[-2] + p[0]) >> 2;

	// Copy back to source

	p = pp;
	dp = dest;

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < w; ++x)
		{
			if(dp[0] < 0)
				p[0] = blur_lookup[0] + orig_lookup[p[0]];
			else if(dp[0] > 255)
				p[0] = blur_lookup[255] + orig_lookup[p[0]];
			else
				p[0] = blur_lookup[dp[0]] + orig_lookup[p[0]];
			p += 2;
			++dp;
		}
		p += diff;
	}
}

#ifdef AVISYNTH_25
void BlurProcessor::ProcessBlock_YV12(unsigned char* p, int pitch, int row_size,
									  int height, int /* block_x */, int /* block_y */,
									  IScriptEnvironment* /* env */)
{
	unsigned char* pp = p;
	short* dp = dest;

	const int h = height - 2, rs = row_size - 2, diff = pitch - row_size;

	// Apply [ 0 0 0, 0 1 1, 0 1 1 ] to top-left corner

	*dp++ = (p[0] + p[1] + p[pitch] + p[1 + pitch]) >> 2;
	++p;

	// Apply [ 0 0 0, 1 1 1, 1 1 1 ] to top row

	for(int x = 0; x < rs; ++x)
	{
		*dp++ = (p[-1] + p[0] + p[1] + p[pitch - 1] + p[pitch] + p[pitch + 1]) / 6;
		++p;
	}

	// Apply [ 0 0 0, 1 1 0, 1 1 0 ] to top-right corner

	*dp++ = (p[-1] + p[0] + p[pitch - 1] + p[pitch]) >> 2;
	++p;

	// Skip "leftover"

	p += diff;

	// y=1 to y=height-1

	for(int y = 0; y < h; ++y)
	{
		// Apply [ 0 1 1, 0 1 1, 0 1 1 ] to left edge

		*dp++ = (p[-pitch] + p[1 - pitch] + p[0] + p[1] + p[pitch] + p[1 + pitch]) / 6;
		++p;

		// Apply [ 1 1 1, 1 1 1, 1 1 1 ]

		for(int x = 0; x < rs; ++x)
		{
			*dp++ = (p[-1 - pitch] + p[-pitch] + p[1 - pitch] + p[-1] + p[0] +
				p[1] + p[pitch - 1] + p[pitch] + p[pitch + 1]) / 9;
			++p;
		}

		// Apply [ 1 1 0, 1 1 0, 1 1 0 ] to right edge

		*dp++ = (p[-1 - pitch] + p[-pitch] + p[-1] + p[0] +
			p[pitch - 1] + p[pitch]) / 6;
		++p;

		// Skip "leftover"

		p += diff;
	}

	// Apply [ 0 1 1, 0 1 1, 0 0 0 ] to bottom-left corner

	*dp++ = (p[-pitch] + p[1 - pitch] + p[0] + p[1]) >> 2;
	++p;

	// Apply [ 1 1 1, 1 1 1, 0 0 0 ] to bottom row

	for(int x = 0; x < rs; ++x)
	{
		*dp++ = (p[-1 - pitch] + p[-pitch] + p[1 - pitch] +
			p[-1] + p[0] + p[1]) / 6;
		++p;
	}

	// Apply [ 1 1 0, 1 1 0, 0 0 0 ] to bottom-right corner

	dp[0] = (p[-1 - pitch] + p[-pitch] + p[-1] + p[0]) >> 2;

	// Copy back to source

	p = pp;
	dp = dest;

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; ++x)
		{
			if(dp[0] < 0)
				*p++ = blur_lookup[0] + orig_lookup[p[0]];
			else if(dp[0] > 255)
				*p++ = blur_lookup[255] + orig_lookup[p[0]];
			else
				*p++ = blur_lookup[dp[0]] + orig_lookup[p[0]];
			++dp;
		}
		p += diff;
	}
}
#endif
