
// Blockbuster
// Avisynth filter for eliminating DCT blocks
//
// By Ross Thomas <ross@grinfinity.com>
//
// There is no copyright on this code, and there are no conditions
// on its distribution or use. Do with it what you will.

// This method applies a Gaussian dither to the block. This differs
// from the noise method in that it applies static noise, i.e. the
// noise is the same for each frame.

#include "Processor.h"
#include "grand.h"

class DitherProcessor : public Processor
{
public:
	DitherProcessor(int block_size, int detail_min,
		int detail_max, int luma_offset, int luma_threshold,
		int frame_width, int frame_height,
		double mean, double variance, int seed,
		IScriptEnvironment* env);
	virtual ~DitherProcessor();

protected:
	virtual void ProcessBlock_YUY2(unsigned char* p, int pitch, int row_size,
		int height, int block_x, int block_y, IScriptEnvironment* env);

#ifdef AVISYNTH_25
	virtual void ProcessBlock_YV12(unsigned char* p, int pitch, int row_size,
		int height, int block_x, int block_y, IScriptEnvironment* env);
#endif

private:
	int frame_width;
	char* buf;
};

DitherProcessor::DitherProcessor(int block_size, int detail_min,
								 int detail_max, int luma_offset,
								 int luma_threshold,
								 int _frame_width, int frame_height,
								 double mean, double variance, int seed,
								 IScriptEnvironment* env)
	: Processor(block_size, detail_min, detail_max,
				luma_offset, luma_threshold, env),
	  frame_width(_frame_width)
{
	if(seed < 0)
		env->ThrowError("Blockbuster: Seed must be >= 0");

	int area = frame_width * frame_height;
	char* b = buf = new char[area];

	GaussianRandom r(seed);

	for(int i = 0; i < area; ++i)
	{
		double d = r.rand(mean, variance);

		if(d < 0.0) d -= 0.5;
		else d += 0.5;

		if(d < -128.0) *b++ = -128;
		else if(d > 127.0) *b++ = 127;
		else *b++ = char(d);
	}
}

DitherProcessor::~DitherProcessor()
{
	delete [] buf;
}

void DitherProcessor::ProcessBlock_YUY2(unsigned char* p, int pitch, int row_size,
										int height, int block_x, int block_y,
										IScriptEnvironment* /* env */)
{
	const char* b = buf + (block_size * block_y) * frame_width + (block_size * block_x);

	int skip = pitch - row_size;
	row_size >>= 1;

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; ++x)
		{
			int n = *p + b[x];
			*p++ = unsigned char(MINMAX(n, 0, 255));
			++p;
		}
		p += skip;
		b += frame_width;
	}
}

#ifdef AVISYNTH_25
void DitherProcessor::ProcessBlock_YV12(unsigned char* p, int pitch, int row_size,
										int height, int block_x, int block_y,
										IScriptEnvironment* /* env */)
{
	const char* b = buf + (block_size * block_y) * frame_width + (block_size * block_x);

	int skip = pitch - row_size;

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; ++x)
		{
			int n = *p + b[x];
			*p++ = unsigned char(MINMAX(n, 0, 255));
		}
		p += skip;
		b += frame_width;
	}
}
#endif
