
// Blockbuster
// Avisynth filter for eliminating DCT blocks
//
// By Ross Thomas <ross@grinfinity.com>
//
// There is no copyright on this code, and there are no conditions
// on its distribution or use. Do with it what you will.

// Base class for processors

#pragma once

#define MINMAX(x, min, max) ((x) < (min) ? (min) : (x) > (max) ? (max) : (x))

class Processor
{
public:
	Processor(int block_size, int detail_min, int detail_max,
		int luma_offset, int luma_threshold, IScriptEnvironment* env);

	void DoFrame_YUY2(unsigned char* p, int pitch, int row_size,
		int height, IScriptEnvironment* env);

#ifdef AVISYNTH_25
	void DoFrame_YV12(unsigned char* p, int pitch, int row_size,
		int height, IScriptEnvironment* env);
#endif

protected:
	virtual void ProcessBlock_YUY2(unsigned char* p, int pitch,
		int row_size, int height, int block_x, int block_y,
		IScriptEnvironment* env) = 0;

#ifdef AVISYNTH_25
	virtual void ProcessBlock_YV12(unsigned char* p, int pitch,
		int row_size, int height, int block_x, int block_y,
		IScriptEnvironment* env) = 0;
#endif

private:
	void DetailCheck_YUY2(unsigned char* p, int pitch, int row_size,
		int height, int block_x, int block_y, IScriptEnvironment* env);
	void AdjustLuma_YUY2(unsigned char* p, int pitch, int row_size,
		int height, IScriptEnvironment* env);

#ifdef AVISYNTH_25
	void DetailCheck_YV12(unsigned char* p, int pitch, int row_size,
		int height, int block_x, int block_y, IScriptEnvironment* env);
	void AdjustLuma_YV12(unsigned char* p, int pitch, int row_size,
		int height, IScriptEnvironment* env);
#endif

protected:
	int block_size;

private:
	int detail_min, detail_max;
	int luma_offset, luma_threshold;
};

Processor::Processor(int _block_size, int _detail_threshold_min,
					 int _detail_threshold_max, int _luma_offset,
					 int _luma_threshold, IScriptEnvironment* env)
	: block_size(_block_size), detail_min(_detail_threshold_min),
	  detail_max(_detail_threshold_max), luma_offset(_luma_offset),
	  luma_threshold(_luma_threshold)
{
	if(block_size < 3)
		env->ThrowError("Blockbuster: Block size must be >= 3");

	if(detail_min < 1 || detail_min > 100)
		env->ThrowError("Blockbuster: detail_min must be >= 1 and <= 100");

	if(detail_max < 1 || detail_max > 100)
		env->ThrowError("Blockbuster: detail_max must be >= 1 and <= 100");

	if(detail_max < detail_min)
		env->ThrowError("Blockbuster: detail_max must be < detail_min");

	if(luma_offset < -255 || luma_offset > 255)
		env->ThrowError("Blockbuster: luma_offset must be >= -255 and <= 255");

	if(luma_threshold < 1 || luma_threshold > 255)
		env->ThrowError("Blockbuster: luma_threshold must be >= 1 and <= 255");

	// Convert from percent to actual count

	detail_min = int(float(block_size * block_size *
		detail_min) / 100.0f + 0.5f);
	detail_max = int(float(block_size * block_size *
		detail_max) / 100.0f + 0.5f);
}

inline void Processor::DoFrame_YUY2(unsigned char* p, int pitch,
									int row_size, int height,
									IScriptEnvironment* env)
{
	const int blocks_across = (row_size >> 1) / block_size,
		block_row_size = block_size << 1,
		remainder_across = (row_size >> 1) % block_size,
		blocks_down = height / block_size,
		remainder_row_size = remainder_across * 2,
		remainder_down = height % block_size,
		end_skip = (pitch - row_size) + (block_size - 1) * pitch;

	for(int y = 0; y < blocks_down; ++y)
	{
		// Do whole blocks

		for(int x = 0; x < blocks_across; ++x)
		{
			DetailCheck_YUY2(p, pitch, block_row_size, block_size, x, y, env);
			p += block_row_size;
		}

		// Do remainder on right

		if(remainder_across > 0)
		{
			DetailCheck_YUY2(p, pitch, remainder_row_size, block_size, x, y, env);
			p += remainder_row_size;
		}

		p += end_skip;
	}

	// Do remainder on bottom

	if(remainder_down > 0)
	{
		for(int x = 0; x < blocks_across; ++x)
		{
			DetailCheck_YUY2(p, pitch, block_row_size, remainder_down, x, y, env);
			p += block_row_size;
		}

		// Don't forget that bit in the corner

		if(remainder_across > 0)
			DetailCheck_YUY2(p, pitch, remainder_row_size, remainder_down, x, y, env);
	}
}

#ifdef AVISYNTH_25
inline void Processor::DoFrame_YV12(unsigned char* p, int pitch, int row_size,
									int height, IScriptEnvironment* env)
{
	const int blocks_across = row_size / block_size,
		remainder_across = row_size % block_size,
		blocks_down = height / block_size,
		remainder_down = height % block_size,
		end_skip = (pitch - row_size) + (block_size - 1) * pitch;

	for(int y = 0; y < blocks_down; ++y)
	{
		// Do whole blocks

		for(int x = 0; x < blocks_across; ++x)
		{
			DetailCheck_YV12(p, pitch, block_size, block_size, x, y, env);
			p += block_size;
		}

		// Do remainder on right

		if(remainder_across > 0)
		{
			DetailCheck_YV12(p, pitch, remainder_across, block_size, x, y, env);
			p += remainder_across;
		}

		p += end_skip;
	}

	// Do remainder on bottom

	if(remainder_down > 0)
	{
		for(int x = 0; x < blocks_across; ++x)
		{
			DetailCheck_YV12(p, pitch, block_size, remainder_down, x, y, env);
			p += block_size;
		}

		// Don't forget that bit in the corner

		if(remainder_across > 0)
			DetailCheck_YV12(p, pitch, remainder_across, remainder_down, x, y, env);
	}
}
#endif

inline void Processor::DetailCheck_YUY2(unsigned char* p, int pitch,
										int row_size, int height,
										int block_x, int block_y,
										IScriptEnvironment* env)
{
	unsigned char* pp = p;
	static unsigned char seen[256];
	int unique = 0;

	memset(seen, 0, sizeof(seen));

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; x += 2)
		{
			if(0 == seen[p[x]]++ && ++unique > detail_max)
				return;
		}
		p += pitch;
	}

	if(unique >= detail_min) // We know it's <= max
	{
		if(luma_offset != 0)
			AdjustLuma_YUY2(pp, pitch, row_size, height, env);
		ProcessBlock_YUY2(pp, pitch, row_size, height, block_x, block_y, env);
	}
}

#ifdef AVISYNTH_25
inline void Processor::DetailCheck_YV12(unsigned char* p, int pitch,
										int row_size, int height,
										int block_x, int block_y,
										IScriptEnvironment* env)
{
	unsigned char* pp = p;
	static unsigned char seen[256];
	int unique = 0;

	memset(seen, 0, sizeof(seen));

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; ++x)
		{
			if(0 == seen[p[x]]++ && ++unique > detail_max)
				return;
		}
		p += pitch;
	}

	if(unique >= detail_min) // We know it's <= max
	{
		if(luma_offset != 0)
			AdjustLuma_YV12(pp, pitch, row_size, height, env);
		ProcessBlock_YV12(pp, pitch, row_size, height, block_x, block_y, env);
	}
}
#endif

inline void Processor::AdjustLuma_YUY2(unsigned char* p, int pitch, int row_size,
									   int height, IScriptEnvironment* /* env */)
{
	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; x += 2)
		{
			if(p[x] <= luma_threshold)
			{
				int n = p[x] + luma_offset;
				p[x] = unsigned char(MINMAX(n, 0, 255));
			}
		}
		p += pitch;
	}
}

#ifdef AVISYNTH_25
inline void Processor::AdjustLuma_YV12(unsigned char* p, int pitch, int row_size,
									   int height, IScriptEnvironment* /* env */)
{
	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; ++x)
		{
			if(p[x] <= luma_threshold)
			{
				int n = p[x] + luma_offset;
				p[x] = unsigned char(MINMAX(n, 0, 255));
			}
		}
		p += pitch;
	}
}
#endif
