
// Blockbuster
// Avisynth filter for eliminating DCT blocks
//
// By Ross Thomas <ross@grinfinity.com>
//
// There is no copyright on this code, and there are no conditions
// on its distribution or use. Do with it what you will.

// This method applies a strong sharpening matrix:
//
//  0 -1  0
// -1  5 -1
//  0 -1  0
//
// to the entire block, except for the edges which use modified
// matrices. The sharpened result is then weighted with the orignal
// based on the strength setting provided.

#include "Processor.h"

class SharpenProcessor : public Processor
{
public:
	SharpenProcessor(int block_size, int detail_min,
		int detail_max, int luma_offset, int luma_threshold,
		int strength, IScriptEnvironment* env);
	virtual ~SharpenProcessor();

protected:
	virtual void Reset() {}

	virtual void ProcessBlock_YUY2(unsigned char* p, int pitch, int row_size,
		int height, int block_x, int block_y, IScriptEnvironment* env);

#ifdef AVISYNTH_25
	virtual void ProcessBlock_YV12(unsigned char* p, int pitch, int row_size,
		int height, int block_x, int block_y, IScriptEnvironment* env);
#endif

private:
	unsigned char sharp_lookup[256], orig_lookup[256];
	short* dest;
};

SharpenProcessor::SharpenProcessor(int block_size, int detail_min,
								   int detail_max, int luma_offset,
								   int luma_threshold, int strength,
								   IScriptEnvironment* env)
	: Processor(block_size, detail_min, detail_max,
				luma_offset, luma_threshold, env)
{
	if(strength < 1 || strength > 100)
		env->ThrowError("Blockbuster: Strength must be >= 1 and <= 100");

	// Generate strength lookup tables

	float sharp_adj = strength / 100.0f, orig_adj = (100 - strength) / 100.0f;
	for(int i = 0; i < 256; ++i)
	{
		sharp_lookup[i] = unsigned char(i * sharp_adj + 0.5f);
		orig_lookup[i] = unsigned char(i * orig_adj + 0.5f);
	}

	// Temp buffer

	dest = new short[block_size * block_size];
}

SharpenProcessor::~SharpenProcessor()
{
	delete [] dest;
}

void SharpenProcessor::ProcessBlock_YUY2(unsigned char* p, int pitch, int row_size,
										 int height, int /* block_x */, int /* block_y */,
										 IScriptEnvironment* /* env */)
{
	unsigned char* pp = p;
	short* dp = dest;
	const int w = row_size / 2, h = height - 2, rs = w - 2,
		diff = pitch - row_size;

	// Apply [ 0 0 0, 0 3 -1, 0 -1 0 ] to top-left corner

	*dp++ = short(-(p[2] + p[pitch]) + 3 * p[0]);
	p += 2;

	// Apply [ 0 0 0, -1 4 -1, 0 -1 0 ] to top row

	for(int x = 0; x < rs; ++x)
	{
		*dp++ = short(-(p[-2] + p[2] + p[pitch]) + 4 * p[0]);
		p += 2;
	}

	// Apply [ 0 0 0, -1 3 0, 0 -1 0 ] to top-right corner

	*dp++ = short(-(p[-2] + p[pitch]) + 3 * p[0]);
	p += 2;

	// Skip "leftover"

	p += diff;

	// y=1 to y=height-1

	for(int y = 0; y < h; ++y)
	{
		// Apply [ 0 -1 0, 0 4 -1, 0 -1 0 ] to left edge

		*dp++ = short(-(p[-pitch] + p[2] + p[pitch]) + 4 * p[0]);
		p += 2;

		// Apply [ 0 -1 0, -1 5 -1, 0 -1 0 ]

		for(int x = 0; x < rs; ++x)
		{
			*dp++ = short(-(p[-pitch] + p[-2] + p[2] + p[pitch]) + 5 * p[0]);
			p += 2;
		}

		// Apply [ 0 -1 0, -1 4 0, 0 -1 0 ] to right edge

		*dp++ = short(-(p[-pitch] + p[-2] + p[pitch]) + 4 * p[0]);
		p += 2;

		// Skip "leftover"

		p += diff;
	}

	// Apply [ 0 -1 0, 0 3 -1, 0 0 0 ] to bottom-left corner

	*dp++ = short(-(p[-pitch] + p[2]) + 3 * p[0]);
	p += 2;

	// Apply [ 0 -1 0, -1 4 -1, 0 0 0 ] to bottom row

	for(int x = 0; x < rs; ++x)
	{
		*dp++ = short(-(p[-pitch] + p[-2] + p[2]) + 4 * p[0]);
		p += 2;
	}

	// Apply [ 0 -1 0, -1 3 0, 0 0 0 ] to bottom-right corner

	dp[0] = short(-(p[-pitch] + p[-2]) + 3 * p[0]);

	// Copy back to source

	p = pp;
	dp = dest;

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < w; ++x)
		{
			if(dp[0] < 0)
				p[0] = sharp_lookup[0] + orig_lookup[p[0]];
			else if(dp[0] > 255)
				p[0] = sharp_lookup[255] + orig_lookup[p[0]];
			else
				p[0] = sharp_lookup[dp[0]] + orig_lookup[p[0]];
			p += 2;
			++dp;
		}
		p += diff;
	}
}

#ifdef AVISYNTH_25
void SharpenProcessor::ProcessBlock_YV12(unsigned char* p, int pitch, int row_size,
										 int height, int /* block_x */, int /* block_y */,
										 IScriptEnvironment* /* env */)
{
	unsigned char* pp = p;
	short* dp = dest;
	const int h = height - 2, rs = row_size - 2, diff = pitch - row_size;

	// Apply [ 0 0 0, 0 3 -1, 0 -1 0 ] to top-left corner

	*dp++ = short(-(p[1] + p[pitch]) + 3 * p[0]);
	++p;

	// Apply [ 0 0 0, -1 4 -1, 0 -1 0 ] to top row

	for(int x = 0; x < rs; ++x)
	{
		*dp++ = short(-(p[-1] + p[1] + p[pitch]) + 4 * p[0]);
		++p;
	}

	// Apply [ 0 0 0, -1 3 0, 0 -1 0 ] to top-right corner

	*dp++ = short(-(p[-1] + p[pitch]) + 3 * p[0]);
	++p;

	// Skip "leftover"

	p += diff;

	// y=1 to y=height-1

	for(int y = 0; y < h; ++y)
	{
		// Apply [ 0 -1 0, 0 4 -1, 0 -1 0 ] to left edge

		*dp++ = short(-(p[-pitch] + p[1] + p[pitch]) + 4 * p[0]);
		++p;

		// Apply [ 0 -1 0, -1 5 -1, 0 -1 0 ]

		for(int x = 0; x < rs; ++x)
		{
			*dp++ = short(-(p[-pitch] + p[-1] + p[1] + p[pitch]) + 5 * p[0]);
			++p;
		}

		// Apply [ 0 -1 0, -1 4 0, 0 -1 0 ] to right edge

		*dp++ = short(-(p[-pitch] + p[-1] + p[pitch]) + 4 * p[0]);
		++p;

		// Skip "leftover"

		p += diff;
	}

	// Apply [ 0 -1 0, 0 3 -1, 0 0 0 ] to bottom-left corner

	*dp++ = short(-(p[-pitch] + p[1]) + 3 * p[0]);
	++p;

	// Apply [ 0 -1 0, -1 4 -1, 0 0 0 ] to bottom row

	for(int x = 0; x < rs; ++x)
	{
		*dp++ = short(-(p[-pitch] + p[-1] + p[1]) + 4 * p[0]);
		++p;
	}

	// Apply [ 0 -1 0, -1 3 0, 0 0 0 ] to bottom-right corner

	dp[0] = short(-(p[-pitch] + p[-1]) + 3 * p[0]);

	// Copy back to source

	p = pp;
	dp = dest;

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; ++x)
		{
			if(dp[0] < 0)
				p[0] = sharp_lookup[0] + orig_lookup[p[0]];
			else if(dp[0] > 255)
				p[0] = sharp_lookup[255] + orig_lookup[p[0]];
			else
				p[0] = sharp_lookup[dp[0]] + orig_lookup[p[0]];
			++p;
			++dp;
		}
		p += diff;
	}
}
#endif
